@@ -6,7 +6,7 @@ Mojo::APNS - Apple Push Notification Service for Mojolicious
 
 =head1 VERSION
 
-0.02
+0.03
 
 =head1 DESCRIPTION
 
@@ -32,6 +32,11 @@ This module does not support password protected SSL keys.
     badge => 2,
   );
 
+  $apns->on(feedback => sub {
+    my($apns, $feedback) = @_;
+    warn "$feedback->{device} rejected push at $feedback->{ts}";
+  });
+
   $apns->ioloop->start;
 
 =cut
@@ -40,21 +45,39 @@ use feature 'state';
 use Mojo::Base 'Mojo::EventEmitter';
 use Mojo::JSON;
 use Mojo::IOLoop;
+use constant FEEDBACK_RECONNECT_TIMEOUT => 5;
 use constant DEBUG => $ENV{MOJO_APNS_DEBUG} ? 1 : 0;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 =head1 EVENTS
 
 =head2 error
 
-Emitted when an error occur between client and server. This event will also
-get "Timeout" events.
+Emitted when an error occur between client and server.
 
 =head2 drain
 
 Emitted once all messages have been sent to the server.
 
+=head2 feedback
+
+  $self->on(feedback => sub {
+    my($self, $data) = @_;
+    # ...
+  });
+
+This event is emitted once a device has rejected a notification. C<$data> is a
+hash-ref:
+
+  {
+    ts => $rejected_epoch_timestamp,
+    device => $device_token,
+  }
+
+Once you start listening to "feedback" events, a connection will be made to
+Apple's push notification server which will then send data to this callback.
+
 =head1 ATTRIBUTES
 
 =head2 cert
@@ -81,16 +104,56 @@ has cert => '';
 has sandbox => 0;
 
 has ioloop => sub { Mojo::IOLoop->singleton };
+has _feedback_port => 2196;
 has _gateway_port => 2195;
 has _gateway_address => sub {
   $_[0]->sandbox ? 'gateway.sandbox.push.apple.com' : 'gateway.push.apple.com'
 };
 
-sub _address { "$_[0]->{_gateway_address}:$_[0]->{_gateway_port}" } # DEBUG
 sub _json { state $json = Mojo::JSON->new }
 
 =head1 METHODS
 
+=head2 on
+
+Same as L<Mojo::EventEmitter/on>, but will also set up feedback connection if
+the event is L</feedback>.
+
+=cut
+
+sub on {
+  my($self, $event, @args) = @_;
+
+  if($event eq 'feedback' and !$self->{feedback_id}) {
+    $self->_connect(feedback => $self->_connected_to_feedback_deamon_cb);
+  }
+
+  $self->SUPER::on($event => @args);
+}
+
+sub _connected_to_feedback_deamon_cb {
+  my $self = shift;
+  my($bytes, $ts, $device) = ('');
+
+  sub {
+    my($self, $stream) = @_;
+    Scalar::Util::weaken($self);
+    $stream->timeout(0);
+    $stream->on(close => sub {
+      $stream->reactor->timer(FEEDBACK_RECONNECT_TIMEOUT, sub {
+        $self or return;
+        $self->_connect(feedback => $self->_connected_to_feedback_deamon_cb);
+      });
+    });
+    $stream->on(read => sub {
+      $bytes .= $_[1];
+      ($ts, $device, $bytes) = unpack 'N n/a a*', $bytes;
+      warn "[APNS:$device] >>> $ts\n" if DEBUG;
+      $self->emit(feedback => { ts => $ts, device => $device });
+    });
+  };
+}
+
 =head2 send
 
   $self->send($device, $message, %args);
@@ -157,29 +220,34 @@ sub send {
 }
 
 sub _connect {
-  my $self = shift;
+  my($self, $type, $cb) = @_;
+  my $port = $type eq 'gateway' ? $self->_gateway_port : $self->_feedback_port;
+  my @cleanup = map { "${type}_$_" } qw/ id stream /;
 
-  warn "[APNS:@{[$self->_address]}] <<< cert=@{[$self->cert]}\n" if DEBUG;
-  warn "[APNS:@{[$self->_address]}] <<< key=@{[$self->key]}\n" if DEBUG;
+  if(DEBUG) {
+    my $key = join ':', $self->_gateway_address, $port;
+    warn "[APNS:$key] <<< cert=@{[$self->cert]}\n" if DEBUG;
+    warn "[APNS:$key] <<< key=@{[$self->key]}\n" if DEBUG;
+  }
 
   Scalar::Util::weaken($self);
-  $self->{client_id}
+  $self->{$cleanup[0]}
     ||= $self->ioloop->client(
         address => $self->_gateway_address,
-        port => $self->_gateway_port,
+        port => $port,
         tls => 1,
         tls_cert => $self->cert,
         tls_key => $self->key,
         sub {
           my($ioloop, $error, $stream) = @_;
 
-          $error and return $self->emit(error => $error);
-          $self->{stream} = $stream;
-          $stream->on(close => sub { delete $self->{$_} for qw/ client_id stream / });
-          $stream->on(error => sub { $self->emit(error => $_[1]) });
-          $stream->on(drain => sub { @{ $self->{messages} } or $self->emit('drain'); });
-          $stream->on(timeout => sub { $self->emit(error => 'Timeout') });
-          $self->_write(shift @{ $self->{messages} });
+          $error and return $self->emit(error => "$type: $error");
+          $self->{$cleanup[1]} = $stream;
+          $stream->on(close => sub { delete $self->{$_} for @cleanup });
+          $stream->on(error => sub { $self->emit(error => "$type: $_[1]") });
+          $stream->on(drain => sub { $self->emit('drain'); });
+          $stream->on(timeout => sub { delete $self->{$_} for @cleanup });
+          $self->$cb($stream);
         },
       );
 }
@@ -191,14 +259,11 @@ sub _default_handler {
 sub _write {
   my($self, $message) = @_;
 
-  $self->{messages} ||= [];
-
-  if($self->{stream}) {
-    $self->{stream}->write(join '', @$message);
+  if($self->{gateway_stream}) {
+    $self->{gateway_stream}->write(join '', @$message);
   }
   else {
-    push @{ $self->{messages} }, $message;
-    $self->_connect unless $self->{client_id};
+    $self->_connect(gateway => sub { shift->_write($message) }) unless $self->{gateway_id};
   }
 
   $self;
@@ -206,9 +271,14 @@ sub _write {
 
 sub DESTROY {
   my $self = shift;
-  my $client_id = $self->{client_id} or return;
   my $ioloop = $self->ioloop or return;
-  $ioloop->remove($client_id);
+
+  if(my $id = $self->{gateway_id}) {
+    $ioloop->remove($id);
+  }
+  if(my $id = $self->{feedback_id}) {
+    $ioloop->remove($id);
+  }
 }
 
 =head1 AUTHOR
